from jarbas_house.devices import GenericDevice, Bulb
from jarbas_house.settings import EWELINK_API_REGION, EWELINK_PASSWORD, EWELINK_USERNAME, EWELINK_API_KEY, \
    EWELINK_API_TOKEN
from jarbas_house.devices.ewelink.sonoff import Sonoff
from jarbas_house.colors import Color
from threading import Thread
from time import sleep


class EwelinkDevice(GenericDevice):
    _sonoff = Sonoff(EWELINK_USERNAME, EWELINK_PASSWORD, EWELINK_API_REGION, EWELINK_API_KEY, EWELINK_API_TOKEN)

    def __init__(self, host, name="generic ewelink device", raw_data=None, product_model="ewelink"):
        raw_data = raw_data or {"name": name, "description": "uses ewelink app"}
        super().__init__(host, name, raw_data)
        self._product_model = product_model or name

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "device_type": "generic ewelink device",
            "device_id": self.device_id,
            "state": self.is_on,
            "online": self.is_online,
            "raw": self.raw_data
        }

    @property
    def is_online(self):
        return self.raw_data.get("online")

    @property
    def product_model(self):
        return self._product_model

    @property
    def is_on(self):
        return self.raw_data.get("params", {}).get("state", "off") == "on"

    @property
    def device_id(self):
        return self.raw_data.get("deviceid")

    # status change
    def turn_on(self):
        self._sonoff.update_params(self.raw_data, {'state': "on"})
        bucket = self.raw_data
        bucket["params"]["state"] = "on"
        self._raw.append(bucket)

    def turn_off(self):
        self._sonoff.update_params(self.raw_data, {'state': "off"})
        bucket = self.raw_data
        bucket["params"]["state"] = "off"
        self._raw.append(bucket)


class EwelinkBulb(EwelinkDevice, Bulb):
    def __init__(self, host, name="bulb", raw_data=None, product_model="ewelink"):
        super().__init__(host, name, raw_data, product_model=product_model)


class MosquitoKiller(EwelinkBulb):
    def __init__(self, host, name="mosquito killer", raw_data=None, product_model="ewelink"):
        super().__init__(host, name, raw_data, product_model=product_model)
        self.mode = ""
        self._timer = None

    @property
    def color(self):
        if self.is_off:
            return Color.from_name("black")
        return Color.from_name("violet")

    def change_color(self, color="violet"):
        if isinstance(color, Color):
            if color.rgb == (0, 0, 0):
                self.turn_off()
            else:
                if self.is_off:
                    self.turn_on()
                if Color.from_name("violet") != color:
                    print("ERROR: mosquito killer does not support color change")
        else:
            color = Color.from_name(color)
            self.change_color(color)

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "mosquito killer",
            "brightness": self.brightness_255,
            "device_id": self.device_id,
            "state": self.is_on,
            "online": self.is_online,
            "raw": self.raw_data
        }

    @property
    def brightness_255(self):
        """
        Return current brightness 0-255
        """
        return self.raw_data["params"]['channel0']

    def change_brightness(self, value, percent=True):
        self.change_channel(0, value, percent)

    def change_channel(self, n, value, percent=True):
        # 0 - brightness
        # 1 - 4 - seem to do nothing ?
        assert 0 <= n <= 4
        if isinstance(value, str):
            value = int(value)

        if percent and isinstance(value, int):
            value = value * 255 / 100
        elif percent and isinstance(value, float) and 0 <= value <= 1:  # 0.1
            value = value * 255

        value = str(int(value))
        # print("Changing mosquito killer light channel: {n} to value: {value}".format(n=n,
        #                                                                             value=value))
        self._sonoff.update_params(self.raw_data, {'channel' + str(n): value})
        bucket = self.raw_data
        bucket["params"]['channel' + str(n)] = value
        self._raw.append(bucket)

    def blink(self, speed=0):

        self.mode = "blink"
        if self.is_off:
            self.turn_on()

        def cycle():
            while self.mode == "blink":
                self.change_brightness(10)
                sleep(0.5)
                self.change_brightness(100)
                sleep(0.5)

        self._timer = Thread(target=cycle)
        self._timer.setDaemon(True)
        self._timer.start()

    def reset(self):
        self.mode = ""
        self.turn_on()
        self.set_high_brightness()
        self._timer = None


# Product models
# add more to this list when support is added
# refers to productModel field
EWELINK_BRAND_LOOKUP = {
    'LA-1CUI': MosquitoKiller
}


def scan_ewelink():
    s = Sonoff(EWELINK_USERNAME, EWELINK_PASSWORD, EWELINK_API_REGION)
    devices = s.get_devices()
    if devices:
        for d in devices:
            product_model = d.get("productModel", "ewelink")
            name = d.get("name", "generic_device")
            host = d.get("ip")
            clazz = EWELINK_BRAND_LOOKUP.get(product_model) or EwelinkDevice
            yield clazz(host, name, d, product_model=product_model)


def get_device(ip):
    for device in scan_ewelink():
        if device.host == ip:
            return device
    return None


if __name__ == "__main__":
    from pprint import pprint
    from time import sleep

    mosquito = get_device("89.181.81.195")
    if mosquito:
        mosquito.toggle()
        print(mosquito.__class__.__name__)

    exit()
    for d in scan_ewelink():
        pprint(d.as_dict)
        # print(d.name, d.host, d.is_on, d.is_online)



import datetime
from jarbas_house.colors import rgb_to_name


def color_tuple_to_string(rgb):
    # try to convert to an english name
    try:
        return rgb_to_name(rgb)
    except Exception:
        # print e
        return str(rgb)


def date_has_passed(dt):
    delta = dt - datetime.datetime.now()
    return delta.total_seconds() < 0


max_delay = 0x1f


def delay_to_speed(delay):
    # speed is 0-100, delay is 1-31
    # 1st translate delay to 0-30
    delay = delay - 1
    if delay > max_delay - 1:
        delay = max_delay - 1
    if delay < 0:
        delay = 0
    inv_speed = int((delay * 100) / (max_delay - 1))
    speed = 100 - inv_speed
    return speed


def speed_to_delay(speed):
    # speed is 0-100, delay is 1-31
    if speed > 100:
        speed = 100
    if speed < 0:
        speed = 0
    inv_speed = 100 - speed
    delay = int((inv_speed * (max_delay - 1)) / 100)
    # translate from 0-30 to 1-31
    delay = delay + 1
    return delay


def byte_to_percent(byte):
    if byte > 255:
        byte = 255
    if byte < 0:
        byte = 0
    return int((byte * 100)/ 255)


def percent_to_byte(percent):
    if percent > 100:
        percent = 100
    if percent < 0:
        percent = 0
    return int((percent * 255) / 100)



#!/usr/bin/env python

"""
This is a utility for controlling stand-alone Flux WiFi LED light bulbs.
The protocol was reverse-engineered by studying packet captures between a
bulb and the controlling "Magic Home" mobile app.  The code here dealing
with the network protocol is littered with magic numbers, and ain't so pretty.
But it does seem to work!

So far most of the functionality of the apps is available here via the CLI
and/or programmatically.

The classes in this project could very easily be used as an API, and incorporated into a GUI app written
in PyQt, Kivy, or some other framework.

##### Available:
* Discovering bulbs on LAN
* Turning on/off bulb
* Get state information
* Setting "warm white" mode
* Setting single color mode
* Setting preset pattern mode
* Setting custom pattern mode
* Reading timers
* Setting timers

##### Some missing pieces:
* Initial administration to set up WiFi SSID and passphrase/key.
* Remote access administration
* Music-relating pulsing. This feature isn't so impressive on the Magic Home app,
and looks like it might be a bit of work.
"""

import socket
import time
import datetime
import colorsys
import threading
from jarbas_house.devices.magic_home import utils


class PresetPattern(object):
    seven_color_cross_fade = 0x25
    red_gradual_change = 0x26
    green_gradual_change = 0x27
    blue_gradual_change = 0x28
    yellow_gradual_change = 0x29
    cyan_gradual_change = 0x2a
    purple_gradual_change = 0x2b
    white_gradual_change = 0x2c
    red_green_cross_fade = 0x2d
    red_blue_cross_fade = 0x2e
    green_blue_cross_fade = 0x2f
    seven_color_strobe_flash = 0x30
    red_strobe_flash = 0x31
    green_strobe_flash = 0x32
    blue_strobe_flash = 0x33
    yellow_strobe_flash = 0x34
    cyan_strobe_flash = 0x35
    purple_strobe_flash = 0x36
    white_strobe_flash = 0x37
    seven_color_jumping = 0x38

    @staticmethod
    def valid(pattern):
        if pattern < 0x25 or pattern > 0x38:
            return False
        return True

    @staticmethod
    def valtostr(pattern):
        for key, value in list(PresetPattern.__dict__.items()):
            if type(value) is int and value == pattern:
                return key.replace("_", " ").title()
        return None


class BuiltInTimer(object):
    sunrise = 0xA1
    sunset = 0xA2

    @staticmethod
    def valid(byte_value):
        return byte_value == BuiltInTimer.sunrise or byte_value == BuiltInTimer.sunset

    @staticmethod
    def valtostr(pattern):
        for key, value in list(BuiltInTimer.__dict__.items()):
            if type(value) is int and value == pattern:
                return key.replace("_", " ").title()
        return None


class LedTimer(object):
    Mo = 0x02
    Tu = 0x04
    We = 0x08
    Th = 0x10
    Fr = 0x20
    Sa = 0x40
    Su = 0x80
    Everyday = Mo | Tu | We | Th | Fr | Sa | Su
    Weekdays = Mo | Tu | We | Th | Fr
    Weekend = Sa | Su

    @staticmethod
    def dayMaskToStr(mask):
        for key, value in list(LedTimer.__dict__.items()):
            if type(value) is int and value == mask:
                return key
        return None

    def __init__(self, bytes=None):
        if bytes is not None:
            self.fromBytes(bytes)
            return

        the_time = datetime.datetime.now() + datetime.timedelta(hours=1)
        self.setTime(the_time.hour, the_time.minute)
        self.setDate(the_time.year, the_time.month, the_time.day)
        self.setModeTurnOff()
        self.setActive(False)

    def setActive(self, active=True):
        self.active = active

    def isActive(self):
        return self.active

    def isExpired(self):
        # if no repeat mask and datetime is in past, return True
        if self.repeat_mask != 0:
            return False
        elif self.year != 0 and self.month != 0 and self.day != 0:
            dt = datetime.datetime(self.year, self.month, self.day, self.hour, self.minute)
            if utils.date_has_passed(dt):
                return True
        return False

    def setTime(self, hour, minute):
        self.hour = hour
        self.minute = minute

    def setDate(self, year, month, day):
        self.year = year
        self.month = month
        self.day = day
        self.repeat_mask = 0

    def setRepeatMask(self, repeat_mask):
        self.year = 0
        self.month = 0
        self.day = 0
        self.repeat_mask = repeat_mask

    def setModeDefault(self):
        self.mode = "default"
        self.pattern_code = 0
        self.turn_on = True
        self.red = 0
        self.green = 0
        self.blue = 0
        self.warmth_level = 0

    def setModePresetPattern(self, pattern, speed):
        self.mode = "preset"
        self.warmth_level = 0
        self.pattern_code = pattern
        self.delay = utils.speed_to_delay(speed)
        self.turn_on = True

    def setModeColor(self, r, g, b):
        self.mode = "color"
        self.warmth_level = 0
        self.red = r
        self.green = g
        self.blue = b
        self.pattern_code = 0x61
        self.turn_on = True

    def setModeWarmWhite(self, level):
        self.mode = "ww"
        self.warmth_level = utils.percent_to_byte(level)
        self.pattern_code = 0x61
        self.red = 0
        self.green = 0
        self.blue = 0
        self.turn_on = True

    def setModeSunrise(self, startBrightness, endBrightness, duration):
        self.mode = "sunrise"
        self.turn_on = True
        self.pattern_code = BuiltInTimer.sunrise
        self.brightness_start = utils.percent_to_byte(startBrightness)
        self.brightness_end = utils.percent_to_byte(endBrightness)
        self.warmth_level = utils.percent_to_byte(endBrightness)
        self.duration = int(duration)

    def setModeSunset(self, startBrightness, endBrightness, duration):
        self.mode = "sunrise"
        self.turn_on = True
        self.pattern_code = BuiltInTimer.sunset
        self.brightness_start = utils.percent_to_byte(startBrightness)
        self.brightness_end = utils.percent_to_byte(endBrightness)
        self.warmth_level = utils.percent_to_byte(endBrightness)
        self.duration = int(duration)

    def setModeTurnOff(self):
        self.mode = "off"
        self.turn_on = False
        self.pattern_code = 0

    """

    timer are in six 14-byte structs
        f0 0f 08 10 10 15 00 00 25 1f 00 00 00 f0 0f
         0  1  2  3  4  5  6  7  8  9 10 11 12 13 14

        0: f0 when active entry/ 0f when not active
        1: (0f=15) year when no repeat, else 0
        2:  month when no repeat, else 0
        3:  dayofmonth when no repeat, else 0
        4: hour
        5: min
        6: 0
        7: repeat mask, Mo=0x2,Tu=0x04, We 0x8, Th=0x10 Fr=0x20, Sa=0x40, Su=0x80
        8:  61 for solid color or warm, or preset pattern code
        9:  r (or delay for preset pattern)
        10: g
        11: b
        12: warm white level
        13: 0f = off, f0 = on ?
    """

    def fromBytes(self, bytes):
        # utils.dump_bytes(bytes)
        self.red = 0
        self.green = 0
        self.blue = 0
        if bytes[0] == 0xf0:
            self.active = True
        else:
            self.active = False
        self.year = bytes[1] + 2000
        self.month = bytes[2]
        self.day = bytes[3]
        self.hour = bytes[4]
        self.minute = bytes[5]
        self.repeat_mask = bytes[7]
        self.pattern_code = bytes[8]

        if self.pattern_code == 0x00:
            self.mode = "default"
        elif self.pattern_code == 0x61:
            self.mode = "color"
            self.red = bytes[9]
            self.green = bytes[10]
            self.blue = bytes[11]
        elif BuiltInTimer.valid(self.pattern_code):
            self.mode = BuiltInTimer.valtostr(self.pattern_code)
            self.duration = bytes[9]  # same byte as red
            self.brightness_start = bytes[10]  # same byte as green
            self.brightness_end = bytes[11]  # same byte as blue
        elif PresetPattern.valid(self.pattern_code):
            self.mode = "preset"
            self.delay = bytes[9]  # same byte as red
        else:
            self.mode = "unknown"

        self.warmth_level = bytes[12]
        if self.warmth_level != 0:
            self.mode = "ww"

        if bytes[13] == 0xf0:
            self.turn_on = True
        else:
            self.turn_on = False
            self.mode = "off"

    def toBytes(self):
        bytes = bytearray(14)
        if not self.active:
            bytes[0] = 0x0f
            # quit since all other zeros is good
            return bytes

        bytes[0] = 0xf0

        if self.year >= 2000:
            bytes[1] = self.year - 2000
        else:
            bytes[1] = self.year
        bytes[2] = self.month
        bytes[3] = self.day
        bytes[4] = self.hour
        bytes[5] = self.minute
        # what is 6?
        bytes[7] = self.repeat_mask

        if not self.turn_on:
            bytes[13] = 0x0f
            return bytes
        bytes[13] = 0xf0

        bytes[8] = self.pattern_code
        if PresetPattern.valid(self.pattern_code):
            bytes[9] = self.delay
            bytes[10] = 0
            bytes[11] = 0
        elif BuiltInTimer.valid(self.pattern_code):
            bytes[9] = self.duration
            bytes[10] = self.brightness_start
            bytes[11] = self.brightness_end
        else:
            bytes[9] = self.red
            bytes[10] = self.green
            bytes[11] = self.blue
        bytes[12] = self.warmth_level

        return bytes

    def __str__(self):
        txt = ""
        if not self.active:
            return "Unset"

        if self.turn_on:
            txt += "[ON ]"
        else:
            txt += "[OFF]"

        txt += " "

        txt += "{:02}:{:02}  ".format(self.hour, self.minute)

        if self.repeat_mask == 0:
            txt += "Once: {:04}-{:02}-{:02}".format(self.year, self.month, self.day)
        else:
            bits = [LedTimer.Su, LedTimer.Mo, LedTimer.Tu, LedTimer.We, LedTimer.Th, LedTimer.Fr, LedTimer.Sa]
            for b in bits:
                if self.repeat_mask & b:
                    txt += LedTimer.dayMaskToStr(b)
                else:
                    txt += "--"
            txt += "  "

        txt += "  "
        if self.pattern_code == 0x61:
            if self.warmth_level != 0:
                txt += "Warm White: {}%".format(utils.byte_to_percent(self.warmth_level))
            else:
                color_str = utils.color_tuple_to_string((self.red, self.green, self.blue))
                txt += "Color: {}".format(color_str)

        elif PresetPattern.valid(self.pattern_code):
            pat = PresetPattern.valtostr(self.pattern_code)
            speed = utils.delay_to_speed(self.delay)
            txt += "{} (Speed:{}%)".format(pat, speed)

        elif BuiltInTimer.valid(self.pattern_code):
            type = BuiltInTimer.valtostr(self.pattern_code)

            txt += "{} (Duration:{} minutes, Brightness: {}% -> {}%)".format(
                type, self.duration,
                utils.byte_to_percent(self.brightness_start), utils.byte_to_percent(self.brightness_end))

        return txt


class WifiLedBulb(object):
    def __init__(self, ipaddr, port=5577, timeout=5):
        self.ipaddr = ipaddr
        self.port = port
        self.timeout = timeout

        self.protocol = None
        self.rgbwcapable = False
        self.rgbwprotocol = False

        self.raw_state = None
        self._is_on = False
        self._mode = None
        self._socket = None
        self._lock = threading.Lock()
        self._query_len = 0
        self._use_csum = True

        self.connect(2)
        self.update_state()

    @property
    def mode(self):
        return self._mode

    @property
    def warm_white(self):
        if self.protocol == 'LEDENET':
            return self.raw_state[9]
        else:
            return 0

    @property
    def cold_white(self):
        if self.protocol == 'LEDENET':
            return self.raw_state[11]
        else:
            return 0

    @property
    def brightness(self):
        """Return current brightness 0-255.

        For warm white return current led level. For RGB
        calculate the HSV and return the 'value'.
        """
        if self.mode == "ww":
            return int(self.raw_state[9])
        else:
            _, _, v = colorsys.rgb_to_hsv(*self.rgb)
            return v

    def connect(self, retry=0):
        self.close()
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.settimeout(self.timeout)
            self._socket.connect((self.ipaddr, self.port))
        except socket.error:
            if retry < 1:
                return
            self.connect(max(retry - 1, 0))

    def close(self):
        if self._socket is None:
            return
        try:
            self._socket.close()
        except socket.error:
            pass

    def _determine_mode(self, ww_level, pattern_code):
        mode = "unknown"
        if pattern_code in [0x61, 0x62]:
            if self.rgbwcapable:
                mode = "color"
            elif ww_level != 0:
                mode = "ww"
            else:
                mode = "color"
        elif pattern_code == 0x60:
            mode = "custom"
        elif pattern_code == 0x41:
            mode = "color"
        elif PresetPattern.valid(pattern_code):
            mode = "preset"
        elif BuiltInTimer.valid(pattern_code):
            mode = BuiltInTimer.valtostr(pattern_code)
        return mode

    def _determine_query_len(self, retry=2):

        # determine the type of protocol based of first 2 bytes.
        self._send_msg(bytearray([0x81, 0x8a, 0x8b]))
        rx = self._read_msg(2)
        # if any response is recieved, use the default protocol
        if len(rx) == 2:
            self._query_len = 14
            return
        # if no response from default received, next try the original protocol
        self._send_msg(bytearray([0xef, 0x01, 0x77]))
        rx = self._read_msg(2)
        if rx[1] == 0x01:
            self.protocol = 'LEDENET_ORIGINAL'
            self._use_csum = False
            self._query_len = 11
            return
        else:
            self._use_csum = True
        if rx == None and retry > 0:
            self._determine_query_len(max(retry - 1, 0))

    def query_state(self, retry=2, led_type=None):
        if self._query_len == 0:
            self._determine_query_len()

        # default value
        msg = bytearray([0x81, 0x8a, 0x8b])
        # alternative for original protocol
        if self.protocol == 'LEDENET_ORIGINAL' or led_type == 'LEDENET_ORIGINAL':
            msg = bytearray([0xef, 0x01, 0x77])
            led_type = 'LEDENET_ORIGINAL'

        try:
            self.connect()
            self._send_msg(msg)
            rx = self._read_msg(self._query_len)
        except socket.error:
            if retry < 1:
                self._is_on = False
                return
            self.connect()
            return self.query_state(max(retry - 1, 0), led_type)
        if rx is None or len(rx) < self._query_len:
            if retry < 1:
                self._is_on = False
                return rx
            return self.query_state(max(retry - 1, 0), led_type)
        return rx

    def update_state(self, retry=2):
        rx = self.query_state(retry)
        if rx is None or len(rx) < self._query_len:
            self._is_on = False
            return

        # typical response:
        # pos  0  1  2  3  4  5  6  7  8  9 10
        #    66 01 24 39 21 0a ff 00 00 01 99
        #     |  |  |  |  |  |  |  |  |  |  |
        #     |  |  |  |  |  |  |  |  |  |  checksum
        #     |  |  |  |  |  |  |  |  |  warmwhite
        #     |  |  |  |  |  |  |  |  blue
        #     |  |  |  |  |  |  |  green
        #     |  |  |  |  |  |  red
        #     |  |  |  |  |  speed: 0f = highest f0 is lowest
        #     |  |  |  |  <don't know yet>
        #     |  |  |  preset pattern
        #     |  |  off(23)/on(24)
        #     |  type
        #     msg head
        #

        # response from a 5-channel LEDENET controller:
        # pos  0  1  2  3  4  5  6  7  8  9 10 11 12 13
        #    81 25 23 61 21 06 38 05 06 f9 01 00 0f 9d
        #     |  |  |  |  |  |  |  |  |  |  |  |  |  |
        #     |  |  |  |  |  |  |  |  |  |  |  |  |  checksum
        #     |  |  |  |  |  |  |  |  |  |  |  |  color mode (f0 colors were set, 0f whites, 00 all were set)
        #     |  |  |  |  |  |  |  |  |  |  |  cold-white
        #     |  |  |  |  |  |  |  |  |  |  <don't know yet>
        #     |  |  |  |  |  |  |  |  |  warmwhite
        #     |  |  |  |  |  |  |  |  blue
        #     |  |  |  |  |  |  |  green
        #     |  |  |  |  |  |  red
        #     |  |  |  |  |  speed: 0f = highest f0 is lowest
        #     |  |  |  |  <don't know yet>
        #     |  |  |  preset pattern
        #     |  |  off(23)/on(24)
        #     |  type
        #     msg head
        #

        # Devices that don't require a separate rgb/w bit
        if (rx[1] == 0x04 or
                rx[1] == 0x33 or
                rx[1] == 0x81):
            self.rgbwprotocol = True

        # Devices that actually support rgbw
        if (rx[1] == 0x04 or
                rx[1] == 0x25 or
                rx[1] == 0x33 or
                rx[1] == 0x81):
            self.rgbwcapable = True

        # Devices that use an 8-byte protocol
        if (rx[1] == 0x25 or
                rx[1] == 0x27 or
                rx[1] == 0x35):
            self.protocol = "LEDENET"

        # Devices that use the original LEDENET protocol
        if rx[1] == 0x01:
            self.protocol = "LEDENET_ORIGINAL"
            self._use_csum = False

        pattern = rx[3]
        ww_level = rx[9]
        mode = self._determine_mode(ww_level, pattern)
        if mode == "unknown":
            if retry < 1:
                return
            self.update_state(max(retry - 1, 0))
            return
        power_state = rx[2]

        if power_state == 0x23:
            self._is_on = True
        elif power_state == 0x24:
            self._is_on = False
        self.raw_state = rx
        self._mode = mode

    def __str__(self):
        rx = self.raw_state
        mode = self.mode

        pattern = rx[3]
        ww_level = rx[9]
        power_state = rx[2]
        power_str = "Unknown power state"

        if power_state == 0x23:
            power_str = "ON "
        elif power_state == 0x24:
            power_str = "OFF "

        delay = rx[5]
        speed = utils.delay_to_speed(delay)
        if mode == "color":
            red = rx[6]
            green = rx[7]
            blue = rx[8]
            mode_str = "Color: {}".format((red, green, blue))
            if self.rgbwcapable:
                mode_str += " White: {}".format(rx[9])
            else:
                mode_str += " Brightness: {}".format(self.brightness)
        elif mode == "ww":
            mode_str = "Warm White: {}%".format(utils.byte_to_percent(ww_level))
        elif mode == "preset":
            pat = PresetPattern.valtostr(pattern)
            mode_str = "Pattern: {} (Speed {}%)".format(pat, speed)
        elif mode == "custom":
            mode_str = "Custom pattern (Speed {}%)".format(speed)
        elif BuiltInTimer.valid(pattern):
            mode_str = BuiltInTimer.valtostr(pattern)
        else:
            mode_str = "Unknown mode 0x{:x}".format(pattern)
        if pattern == 0x62:
            mode_str += " (tmp)"
        mode_str += " raw state: "
        for _r in rx:
            mode_str += str(_r) + ","
        return "{} [{}]".format(power_str, mode_str)

    def _change_state(self, retry, turn_on=True):

        if self.protocol == 'LEDENET_ORIGINAL':
            msg_on = bytearray([0xcc, 0x23, 0x33])
            msg_off = bytearray([0xcc, 0x24, 0x33])
        else:
            msg_on = bytearray([0x71, 0x23, 0x0f])
            msg_off = bytearray([0x71, 0x24, 0x0f])

        if turn_on:
            msg = msg_on
        else:
            msg = msg_off

        try:
            self._send_msg(msg)
        except socket.error:
            if retry > 0:
                self.connect()
                self._change_state(max(retry - 1, 0), turn_on)
                return
            self._is_on = False

    def turn_on(self, retry=2):
        self._is_on = True
        self._change_state(retry, turn_on=True)

    def turn_off(self, retry=2):
        self._is_on = False
        self._change_state(retry, turn_on=False)

    @property
    def is_on(self):
        return self._is_on

    @property
    def warm_white_255(self):
        if self.mode != "ww":
            return 255
        return self.brightness

    def set_warm_white(self, level, persist=True, retry=2):
        self.set_warm_white_255(utils.percent_to_byte(level), persist, retry)

    def set_warm_white_255(self, level, persist=True, retry=2):
        self.set_rgbw(w=level, persist=persist, brightness=None, retry=retry)

    def set_cold_white(self, level, persist=True, retry=2):
        self.set_cold_white_255(utils.percent_to_byte(level), persist, retry)

    def set_cold_white_255(self, level, persist=True, retry=2):
        self.set_rgbw(persist=persist, brightness=None, retry=retry, w2=level)

    def set_white_temperature(self, temperature, brightness, persist=True,
                              retry=2):
        # Assume output temperature of between 2700 and 6500 Kelvin, and scale
        # the warm and cold LEDs linearly to provide that
        temperature = max(temperature - 2700, 0)
        warm = 255 * (1 - (temperature // 3800))
        cold = min(255 * temperature // 3800, 255)
        warm *= brightness // 255
        cold *= brightness // 255
        self.set_rgbw(w=warm, w2=cold, persist=persist, retry=retry)

    @property
    def rgbw(self):
        if self.mode != "color":
            return (255, 255, 255, 255)
        red = self.raw_state[6]
        green = self.raw_state[7]
        blue = self.raw_state[8]
        white = self.raw_state[9]
        return (red, green, blue, white)

    @property
    def rgbww(self):
        if self.mode != "color":
            return (255, 255, 255, 255, 255)
        red = self.raw_state[6]
        green = self.raw_state[7]
        blue = self.raw_state[8]
        white = self.raw_state[9]
        white2 = self.raw_state[11]
        return (red, green, blue, white, white2)

    def speed(self):
        delay = self.raw_state[5]
        speed = utils.delay_to_speed(delay)
        return speed

    def set_rgbw(self, r=None, g=None, b=None, w=None, persist=True,
                 brightness=None, retry=2, w2=None):

        if (r or g or b) and (w or w2) and not self.rgbwcapable:
            print("RGBW command sent to non-RGBW device")
            raise Exception

        # sample message for original LEDENET protocol (w/o checksum at end)
        #  0  1  2  3  4
        # 56 90 fa 77 aa
        #  |  |  |  |  |
        #  |  |  |  |  terminator
        #  |  |  |  blue
        #  |  |  green
        #  |  red
        #  head

        # sample message for 8-byte protocols (w/ checksum at end)
        #  0  1  2  3  4  5  6
        # 31 90 fa 77 00 00 0f
        #  |  |  |  |  |  |  |
        #  |  |  |  |  |  |  terminator
        #  |  |  |  |  |  write mask / white2 (see below)
        #  |  |  |  |  white
        #  |  |  |  blue
        #  |  |  green
        #  |  red
        #  persistence (31 for true / 41 for false)
        #
        # byte 5 can have different values depending on the type
        # of device:
        # For devices that support 2 types of white value (warm and cold
        # white) this value is the cold white value. These use the LEDENET
        # protocol. If a second value is not given, reuse the first white value.
        #
        # For devices that cannot set both rbg and white values at the same time
        # (including devices that only support white) this value
        # specifies if this command is to set white value (0f) or the rgb
        # value (f0).
        #
        # For all other rgb and rgbw devices, the value is 00

        # sample message for 9-byte LEDENET protocol (w/ checksum at end)
        #  0  1  2  3  4  5  6  7
        # 31 bc c1 ff 00 00 f0 0f
        #  |  |  |  |  |  |  |  |
        #  |  |  |  |  |  |  |  terminator
        #  |  |  |  |  |  |  write mode (f0 colors, 0f whites, 00 colors & whites)
        #  |  |  |  |  |  cold white
        #  |  |  |  |  warm white
        #  |  |  |  blue
        #  |  |  green
        #  |  red
        #  persistence (31 for true / 41 for false)
        #

        if brightness != None:
            (r, g, b) = self._calculate_brightness((r, g, b), brightness)

        # The original LEDENET protocol
        if self.protocol == 'LEDENET_ORIGINAL':
            msg = bytearray([0x56])
            msg.append(int(r))
            msg.append(int(g))
            msg.append(int(b))
            msg.append(0xaa)
        else:
            # all other devices

            # assemble the message
            if persist:
                msg = bytearray([0x31])
            else:
                msg = bytearray([0x41])

            if r is not None:
                msg.append(int(r))
            else:
                msg.append(int(0))
            if g is not None:
                msg.append(int(g))
            else:
                msg.append(int(0))
            if b is not None:
                msg.append(int(b))
            else:
                msg.append(int(0))
            if w is not None:
                msg.append(int(w))
            else:
                msg.append(int(0))

            if self.protocol == "LEDENET":
                # LEDENET devices support two white outputs for cold and warm. We set
                # the second one here - if we're only setting a single white value,
                # we set the second output to be the same as the first
                if w2 is not None:
                    msg.append(int(w2))
                elif w is not None:
                    msg.append(int(w))
                else:
                    msg.append(0)

            # write mask, default to writing color and whites simultaneously
            write_mask = 0x00
            # rgbwprotocol devices always overwrite both color & whites
            if not self.rgbwprotocol:
                if w is None and w2 is None:
                    # Mask out whites
                    write_mask |= 0xf0
                elif r is None and g is None and b is None:
                    # Mask out colors
                    write_mask |= 0x0f

            msg.append(write_mask)

            # Message terminator
            msg.append(0x0f)

        # send the message
        try:
            self._send_msg(msg)
        except socket.error:
            if retry:
                self.connect()
                self.set_rgbw(r, g, b, w, persist=persist, brightness=brightness,
                              retry=max(retry - 1, 0), w2=w2)

    @property
    def rgb(self):
        if self.mode != "color":
            return (255, 255, 255)
        red = self.raw_state[6]
        green = self.raw_state[7]
        blue = self.raw_state[8]
        return (red, green, blue)

    def set_rgb(self, r, g, b, persist=True, brightness=None, retry=2):
        self.set_rgbw(r, g, b, persist=persist, brightness=brightness,
                      retry=retry)

    def _calculate_brightness(self, rgb, level):
        r = rgb[0]
        g = rgb[1]
        b = rgb[2]
        hsv = colorsys.rgb_to_hsv(r, g, b)
        return colorsys.hsv_to_rgb(hsv[0], hsv[1], level)

    def _send_msg(self, bytes):
        # calculate checksum of byte array and add to end
        if self._use_csum:
            csum = sum(bytes) & 0xFF
            bytes.append(csum)
        with self._lock:
            self._socket.send(bytes)

    def _read_msg(self, expected):
        remaining = expected
        rx = bytearray()
        begin = time.time()
        while remaining > 0:
            if time.time() - begin > self.timeout:
                break
            try:
                with self._lock:
                    self._socket.setblocking(0)
                    chunk = self._socket.recv(remaining)
                    if chunk:
                        begin = time.time()
                    remaining -= len(chunk)
                    rx.extend(chunk)
            except socket.error:
                pass
            finally:
                self._socket.setblocking(1)
        return rx

    @property
    def clock(self):
        msg = bytearray([0x11, 0x1a, 0x1b, 0x0f])
        self._send_msg(msg)
        rx = self._read_msg(12)
        if len(rx) != 12:
            return
        year = rx[3] + 2000
        month = rx[4]
        date = rx[5]
        hour = rx[6]
        minute = rx[7]
        second = rx[8]
        # dayofweek = rx[9]
        try:
            dt = datetime.datetime(year, month, date, hour, minute, second)
        except:
            dt = None
        return dt

    def set_clock(self):
        msg = bytearray([0x10, 0x14])
        now = datetime.datetime.now()
        msg.append(now.year - 2000)
        msg.append(now.month)
        msg.append(now.day)
        msg.append(now.hour)
        msg.append(now.minute)
        msg.append(now.second)
        msg.append(now.isoweekday())  # day of week
        msg.append(0x00)
        msg.append(0x0f)
        self._send_msg(msg)

    def set_protocol(self, protocol):
        self.protocol = protocol.upper()

    def set_preset_pattern(self, pattern, speed=80):

        PresetPattern.valtostr(pattern)
        if not PresetPattern.valid(pattern):
            # print "Pattern must be between 0x25 and 0x38"
            raise Exception

        delay = utils.speed_to_delay(speed)
        # print "speed {}, delay 0x{:02x}".format(speed,delay)
        pattern_set_msg = bytearray([0x61])
        pattern_set_msg.append(pattern)
        pattern_set_msg.append(delay)
        pattern_set_msg.append(0x0f)

        self._send_msg(pattern_set_msg)

    @property
    def timers(self):
        msg = bytearray([0x22, 0x2a, 0x2b, 0x0f])
        self._send_msg(msg)
        resp_len = 88
        rx = self._read_msg(resp_len)
        if len(rx) != resp_len:
            print("response too short!")
            raise Exception

        # utils.dump_data(rx)
        start = 2
        timer_list = []
        # pass in the 14-byte timer structs
        for i in range(6):
            timer_bytes = rx[start:][:14]
            timer = LedTimer(timer_bytes)
            timer_list.append(timer)
            start += 14

        return timer_list

    def send_timers(self, timer_list):
        # remove inactive or expired timers from list
        for t in timer_list:
            if not t.isActive() or t.isExpired():
                timer_list.remove(t)

        # truncate if more than 6
        if len(timer_list) > 6:
            print("too many timers, truncating list")
            del timer_list[6:]

        # pad list to 6 with inactive timers
        if len(timer_list) != 6:
            for i in range(6 - len(timer_list)):
                timer_list.append(LedTimer())

        msg_start = bytearray([0x21])
        msg_end = bytearray([0x00, 0xf0])
        msg = bytearray()

        # build message
        msg.extend(msg_start)
        for t in timer_list:
            msg.extend(t.toBytes())
        msg.extend(msg_end)
        self._send_msg(msg)

        # not sure what the resp is, prob some sort of ack?
        rx = self._read_msg(1)
        rx = self._read_msg(3)

    def set_custom_pattern(self, rgb_list, speed, transition_type):
        # truncate if more than 16
        if len(rgb_list) > 16:
            print("too many colors, truncating list")
            del rgb_list[16:]

        # quit if too few
        if len(rgb_list) == 0:
            print("no colors, aborting")
            return

        msg = bytearray()

        first_color = True
        for rgb in rgb_list:
            if first_color:
                lead_byte = 0x51
                first_color = False
            else:
                lead_byte = 0
            r, g, b = rgb
            msg.extend(bytearray([lead_byte, r, g, b]))

        # pad out empty slots
        if len(rgb_list) != 16:
            for i in range(16 - len(rgb_list)):
                msg.extend(bytearray([0, 1, 2, 3]))

        msg.append(0x00)
        msg.append(utils.speed_to_delay(speed))

        if transition_type == "gradual":
            msg.append(0x3a)
        elif transition_type == "jump":
            msg.append(0x3b)
        elif transition_type == "strobe":
            msg.append(0x3c)
        else:
            # unknown transition string: using 'gradual'
            msg.append(0x3a)
        msg.append(0xff)
        msg.append(0x0f)

        self._send_msg(msg)

    def refresh_state(self):
        return self.update_state()


class BulbScanner(object):
    def __init__(self):
        self.found_bulbs = []

    def get_bulb_info_by_id(self, id):
        for b in self.found_bulbs:
            if b['id'] == id:
                return b
        return b

    def get_bulb_info(self):
        return self.found_bulbs

    def scan(self, timeout=10):

        DISCOVERY_PORT = 48899

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind(('', DISCOVERY_PORT))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        msg = "HF-A11ASSISTHREAD".encode('ascii')

        # set the time at which we will quit the search
        quit_time = time.time() + timeout

        response_list = []
        # outer loop for query send
        while True:
            if time.time() > quit_time:
                break
            # send out a broadcast query
            sock.sendto(msg, ('<broadcast>', DISCOVERY_PORT))

            # inner loop waiting for responses
            while True:

                sock.settimeout(1)
                try:
                    data, addr = sock.recvfrom(64)
                except socket.timeout:
                    data = None
                    if time.time() > quit_time:
                        break

                if data is None:
                    continue
                if data == msg:
                    continue

                data = data.decode('ascii')
                data_split = data.split(',')
                if len(data_split) < 3:
                    continue
                item = dict()
                item['ipaddr'] = data_split[0]
                item['id'] = data_split[1]
                item['model'] = data_split[2]
                response_list.append(item)

        self.found_bulbs = response_list
        return response_list




# source: https://github.com/lucien2k/sonoff-python
import logging, time, hmac, hashlib, random, base64, json, socket, requests, re
from datetime import timedelta
from websocket import create_connection

SCAN_INTERVAL = timedelta(seconds=60)

_LOGGER = logging.getLogger(__name__)


def gen_nonce(length=8):
    """Generate pseudorandom number."""
    return ''.join([str(random.randint(0, 9)) for i in range(length)])


class Sonoff:
    def __init__(self, username, password, api_region, user_apikey=None, bearer_token=None, grace_period=600):

        self._username = username
        self._password = password
        self._api_region = api_region
        self._wshost = None

        self._skipped_login = 0
        self._grace_period = timedelta(seconds=grace_period)

        self._user_apikey = user_apikey
        self._bearer_token = bearer_token
        self._devices = []
        self._ws = None

        if user_apikey and bearer_token:
            self.do_reconnect()
        else:
            self.do_login()

    def do_reconnect(self):
        self._headers = {
            'Authorization': 'Bearer ' + self._bearer_token,
            'Content-Type': 'application/json;charset=UTF-8'
        }

        try:
            # get the websocket host
            if not self._wshost:
                self.set_wshost()

            self.update_devices()  # to get the devices list
        except:
            self.do_login()

    def do_login(self):
        import uuid

        # reset the grace period
        self._skipped_login = 0

        app_details = {
            'password': self._password,
            'version': '6',
            'ts': int(time.time()),
            'nonce': gen_nonce(15),
            'appid': 'oeVkj2lYFGnJu5XUtWisfW4utiN4u9Mq',
            'imei': str(uuid.uuid4()),
            'os': 'iOS',
            'model': 'iPhone10,6',
            'romVersion': '11.1.2',
            'appVersion': '3.5.3'
        }

        if re.match(r'[^@]+@[^@]+\.[^@]+', self._username):
            app_details['email'] = self._username
        else:
            app_details['phoneNumber'] = self._username

        decryptedAppSecret = b'6Nz4n0xA8s8qdxQf2GqurZj2Fs55FUvM'

        hex_dig = hmac.new(
            decryptedAppSecret,
            str.encode(json.dumps(app_details)),
            digestmod=hashlib.sha256).digest()

        sign = base64.b64encode(hex_dig).decode()

        self._headers = {
            'Authorization': 'Sign ' + sign,
            'Content-Type': 'application/json;charset=UTF-8'
        }

        r = requests.post('https://{}-api.coolkit.cc:8080/api/user/login'.format(self._api_region),
                          headers=self._headers, json=app_details)

        resp = r.json()
        # get a new region to login
        # TODO error codes
        if 'error' in resp and 'region' in resp:  # and resp['error'] == HTTP_MOVED_PERMANENTLY:
            self._api_region = resp['region']

            _LOGGER.warning(
                "found new region: >>> %s <<< (you should change api_region option to this value in configuration.yaml)",
                self._api_region)

            # re-login using the new localized endpoint
            self.do_login()
            return

        elif 'error' in resp:  # and resp['error'] in [HTTP_NOT_FOUND, HTTP_BAD_REQUEST]:
            # (most likely) login with +86... phone number and region != cn
            if '@' not in self._username and self._api_region != 'cn':
                self._api_region = 'cn'
                self.do_login()

            else:
                _LOGGER.error("Couldn't authenticate using the provided credentials!")

            return
        elif 'error' in resp and resp['error'] == 400:
            # no credentials provided most likely
            _LOGGER.error("Couldn't authenticate using the provided credentials!")

            return

        self._bearer_token = resp['at']
        self._user_apikey = resp['user']['apikey']
        self._headers.update({'Authorization': 'Bearer ' + self._bearer_token})

        # get the websocket host
        if not self._wshost:
            self.set_wshost()

        self.update_devices()  # to get the devices list

    def set_wshost(self):
        r = requests.post('https://%s-disp.coolkit.cc:8080/dispatch/app' % self._api_region, headers=self._headers)
        resp = r.json()

        if 'error' in resp and resp['error'] == 0 and 'domain' in resp:
            self._wshost = resp['domain']
            _LOGGER.info("Found websocket address: %s", self._wshost)
        else:
            raise Exception('No websocket domain')

    def is_grace_period(self):
        grace_time_elapsed = self._skipped_login * int(SCAN_INTERVAL.total_seconds())
        grace_status = grace_time_elapsed < int(self._grace_period.total_seconds())

        if grace_status:
            self._skipped_login += 1

        return grace_status

    def update_devices(self):

        # the login failed, nothing to update
        if not self._wshost:
            return []

        # we are in the grace period, no updates to the devices
        if self._skipped_login and self.is_grace_period():
            _LOGGER.info("Grace period active")
            return self._devices

        r = requests.get('https://{}-api.coolkit.cc:8080/api/user/device'.format(self._api_region),
                         headers=self._headers)

        resp = r.json()
        if 'error' in resp:  # and resp['error'] in [HTTP_BAD_REQUEST, HTTP_UNAUTHORIZED]:
            # @IMPROVE add maybe a service call / switch to deactivate sonoff component
            if self.is_grace_period():
                _LOGGER.warning("Grace period activated!")

                # return the current (and possible old) state of devices
                # in this period any change made with the mobile app (on/off) won't be shown in HA
                return self._devices

            _LOGGER.info("Re-login component")
            self.do_login()

        self._devices = r.json()
        return self._devices

    def get_devices(self, force_update=False):
        if force_update:
            return self.update_devices()

        return self._devices

    def get_device(self, deviceid):
        for device in self.get_devices():
            if 'deviceid' in device and device['deviceid'] == deviceid:
                return device

    def get_api_region(self):
        return self._api_region

    def get_bearer_token(self):
        return self._bearer_token

    def get_user_apikey(self):
        return self._user_apikey

    def _get_ws(self):
        """Check if the websocket is setup and connected."""

        if self._ws is None:
            try:
                self._ws = create_connection(('wss://{}:8080/api/ws'.format(self._wshost)), timeout=10)

                payload = {
                    'action': "userOnline",
                    'userAgent': 'app',
                    'version': 6,
                    'nonce': gen_nonce(15),
                    'apkVesrion': "1.8",
                    'os': 'ios',
                    'at': self.get_bearer_token(),
                    'apikey': self.get_user_apikey(),
                    'ts': str(int(time.time())),
                    'model': 'iPhone10,6',
                    'romVersion': '11.1.2',
                    'sequence': str(time.time()).replace('.', '')
                }

                self._ws.send(json.dumps(payload))
                wsresp = self._ws.recv()
                # _LOGGER.error("open socket: %s", wsresp)

            except (socket.timeout, ConnectionRefusedError, ConnectionResetError):
                _LOGGER.error('failed to create the websocket')
                self._ws = None

        return self._ws

    def update_params(self, device, params):
        self._ws = self._get_ws()

        if not self._ws:
            _LOGGER.warning('invalid websocket, state cannot be changed')
            return
        payload = {
            'action': 'update',
            'userAgent': 'app',
            'params': params,
            'apikey': device['apikey'],
            'deviceid': str(device['deviceid']),
            'sequence': str(time.time()).replace('.', ''),
            'controlType': device['params']['controlType'] if 'controlType' in device['params'] else 4,
            'ts': 0
        }

        # this key is needed for a shared device
        if device['apikey'] != self.get_user_apikey():
            payload['selfApikey'] = self.get_user_apikey()

        self._ws.send(json.dumps(payload))
        wsresp = self._ws.recv()
        # _LOGGER.debug("switch socket: %s", wsresp)

        self._ws.close()  # no need to keep websocket open (for now)
        self._ws = None

    def _device(self, deviceid):
        # we're in the grace period, no state change
        if self._skipped_login:
            _LOGGER.info("Grace period, no state change")
            return

        device = self.get_device(deviceid)
        return device

    def state(self, new_state, deviceid):
        """Switch on or off."""
        device = self._device(deviceid)

        if not device:
            _LOGGER.error('unknown device to be updated')
            return False

        # convert from True/False to on/off
        if isinstance(new_state, (bool)):
            new_state = 'on' if new_state else 'off'

        _LOGGER.debug("Switching `%s` to state: %s", deviceid, new_state)

        params = {'state': new_state}

        self.update_params(device, params)

        # set also te pseudo-internal state of the device until the real refresh kicks in
        for idx, device in enumerate(self._devices):
            if device['deviceid'] == deviceid:
                self._devices[idx]['params']['state'] = new_state

        # @TODO add some sort of validation here, maybe call the devices status
        # only IF MAIN STATUS is done over websocket exclusively

        return new_state

    def switch(self, new_state, deviceid, outlet):
        """Switch on or off."""
        device = self._device(deviceid)

        if not device:
            _LOGGER.error('unknown device to be updated')
            return False

        # convert from True/False to on/off
        if isinstance(new_state, (bool)):
            new_state = 'on' if new_state else 'off'

        if outlet is not None:
            _LOGGER.debug("Switching `%s - %s` on outlet %d to state: %s", \
                          device['deviceid'], device['name'], (outlet + 1), new_state)
        else:
            _LOGGER.debug("Switching `%s` to state: %s", deviceid, new_state)

        # the payload rule is like this:
        #   normal device (non-shared)
        #       apikey      = login apikey (= device apikey too)
        #
        #   shared device
        #       apikey      = device apikey
        #       selfApiKey  = login apikey (yes, it's typed corectly selfApikey and not selfApiKey :|)

        if outlet is not None:
            params = {'switches': device['params']['switches']}
            params['switches'][outlet]['switch'] = new_state

        else:
            params = {'switch': new_state}

        self.update_params(device, params)

        # set also te pseudo-internal state of the device until the real refresh kicks in
        for idx, device in enumerate(self._devices):
            if device['deviceid'] == deviceid:
                if outlet is not None:
                    self._devices[idx]['params']['switches'][outlet]['switch'] = new_state
                else:
                    self._devices[idx]['params']['switch'] = new_state

        # @TODO add some sort of validation here, maybe call the devices status
        # only IF MAIN STATUS is done over websocket exclusively

        return new_state



from jarbas_house.devices import GenericDevice, RGBWBulb, RGBBulb, Bulb
from jarbas_house.devices.magic_home.magichome import BulbScanner, WifiLedBulb, PresetPattern
from jarbas_house.colors import Color


class MagicHomeDevice(GenericDevice):
    def __init__(self, host, name="generic magic_home device", raw_data=None, product_model="magic_home"):
        raw_data = raw_data or {"name": name, "description": "uses magic home app"}
        super().__init__(host, name, raw_data)
        self._product_model = product_model or name

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "generic magic_home device",
            "device_id": self.device_id,
            "state": self.is_on,
            "online": self.is_online,
            "raw": self.raw_data
        }

    @property
    def product_model(self):
        return self._product_model

    @property
    def is_online(self):
        return self.raw_data.get("online", True)

    @property
    def device_id(self):
        return self.raw_data.get("id")


class MagicHomeBulb(MagicHomeDevice, Bulb):

    def __init__(self, host, name="light bulb", raw_data=None, product_model="magic_home"):
        super().__init__(host, name, raw_data, product_model=product_model)
        self._bulb = WifiLedBulb(self.host)
        self._timer = None

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "magic_home light bulb",
            "device_id": self.device_id,
            "brightness": self.brightness_255,
            "state": self.is_on,
            "online": self.is_online,
            "raw": self.raw_data
        }

    @property
    def is_on(self):
        return self._bulb.is_on

    # status change
    def turn_on(self):
        self._bulb.turn_on()

    def turn_off(self):
        self._bulb.turn_off()


class MagicHomeRGBBulb(MagicHomeBulb, RGBBulb):

    def __init__(self, host, name="rgb light bulb", raw_data=None, product_model="magic_home"):
        super().__init__(host, name, raw_data, product_model=product_model)
        self._bulb = WifiLedBulb(self.host)

    def set_preset_pattern(self, pattern, speed=80):
        self._bulb.set_preset_pattern(pattern, speed)

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "magic_home rgb light bulb",
            "color": self.color.as_dict,
            "brightness": self.brightness_255,
            "device_id": self.device_id,
            "state": self.is_on,
            "online": self.is_online,
            "raw": self.raw_data
        }

    @property
    def color(self):
        if self.is_off:
            return Color.from_name("black")
        return Color(*self._bulb.rgb)

    def change_brightness(self, value, percent=True):
        self._bulb.set_rgb(*self.color.rgb, brightness=value)

    def change_color(self, color):
        if isinstance(color, Color):
            color = color.rgb
            if color == (0, 0, 0):
                self.turn_off()
            else:
                if self.is_off:
                    self.turn_on()
                self._bulb.set_rgb(*color)
        else:
            color = Color.from_name(color)
            self.change_color(color)

    @property
    def preset_patterns(self):
        return {
            "seven_color_cross_fade": PresetPattern.seven_color_cross_fade,
            "red_gradual_change": PresetPattern.red_gradual_change,
            "green_gradual_change": PresetPattern.green_gradual_change,
            "blue_gradual_change": PresetPattern.blue_gradual_change,
            "yellow_gradual_change": PresetPattern.yellow_gradual_change,
            "cyan_gradual_change": PresetPattern.cyan_gradual_change,
            "purple_gradual_change": PresetPattern.purple_gradual_change,
            "white_gradual_change": PresetPattern.white_gradual_change,
            "red_green_cross_fade": PresetPattern.red_green_cross_fade,
            "red_blue_cross_fade": PresetPattern.red_blue_cross_fade,
            "green_blue_cross_fade": PresetPattern.green_blue_cross_fade,
            "seven_color_strobe_flash": PresetPattern.seven_color_strobe_flash,
            "red_strobe_flash": PresetPattern.red_strobe_flash,
            "green_strobe_flash": PresetPattern.green_strobe_flash,
            "blue_strobe_flash": PresetPattern.blue_strobe_flash,
            "yellow_strobe_flash": PresetPattern.yellow_strobe_flash,
            "cyan_strobe_flash": PresetPattern.cyan_strobe_flash,
            "purple_strobe_flash": PresetPattern.purple_strobe_flash,
            "white_strobe_flash": PresetPattern.white_strobe_flash,
            "seven_color_jumping": PresetPattern.seven_color_jumping
        }

    def seven_color_cross_fade(self):
        self.set_preset_pattern(self.preset_patterns["seven_color_cross_fade"])

    def red_gradual_change(self):
        self.set_preset_pattern(self.preset_patterns["red_gradual_change"])

    def green_gradual_change(self):
        self.set_preset_pattern(self.preset_patterns["green_gradual_change"])

    def blue_gradual_change(self):
        self.set_preset_pattern(self.preset_patterns["blue_gradual_change"])

    def yellow_gradual_change(self):
        self.set_preset_pattern(self.preset_patterns["yellow_gradual_change"])

    def cyan_gradual_change(self):
        self.set_preset_pattern(self.preset_patterns["cyan_gradual_change"])

    def purple_gradual_change(self):
        self.set_preset_pattern(self.preset_patterns["purple_gradual_change"])

    def white_gradual_change(self):
        self.set_preset_pattern(self.preset_patterns["white_gradual_change"])

    def red_green_cross_fade(self):
        self.set_preset_pattern(self.preset_patterns["red_green_cross_fade"])

    def red_blue_cross_fade(self):
        self.set_preset_pattern(self.preset_patterns["red_blue_cross_fade"])

    def green_blue_cross_fade(self):
        self.set_preset_pattern(self.preset_patterns["green_blue_cross_fade"])

    def seven_color_strobe_flash(self):
        self.set_preset_pattern(self.preset_patterns["seven_color_strobe_flash"])

    def red_strobe_flash(self):
        self.set_preset_pattern(self.preset_patterns["red_strobe_flash"])

    def green_strobe_flash(self):
        self.set_preset_pattern(self.preset_patterns["green_strobe_flash"])

    def blue_strobe_flash(self):
        self.set_preset_pattern(self.preset_patterns["blue_strobe_flash"])

    def yellow_strobe_flash(self):
        self.set_preset_pattern(self.preset_patterns["yellow_strobe_flash"])

    def cyan_strobe_flash(self):
        self.set_preset_pattern(self.preset_patterns["cyan_strobe_flash"])

    def purple_strobe_flash(self):
        self.set_preset_pattern(self.preset_patterns["purple_strobe_flash"])

    def white_strobe_flash(self):
        self.set_preset_pattern(self.preset_patterns["white_strobe_flash"])

    def seven_color_jumping(self):
        self.set_preset_pattern(self.preset_patterns["seven_color_jumping"])


class MagicHomeRGBWBulb(MagicHomeRGBBulb, RGBWBulb):

    def __init__(self, host, name="rgbw light bulb", raw_data=None, product_model="magic_home"):
        super().__init__(host, name, raw_data, product_model=product_model)

    def reset(self):
        self.mode = ""
        self._timer = None
        if self.is_off:
            self.turn_on()
            from time import sleep
            sleep(0.5)
        self.set_warm_white_255(255)

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "magic_home rgbw light bulb",
            "color": self.color.as_dict,
            "brightness": self.brightness_255,
            "device_id": self.device_id,
            "state": self.is_on,
            "online": self.is_online,
            "raw": self.raw_data
        }

    def set_warm_white(self, level, persist=True, retry=2):
        self._bulb.set_warm_white(level, persist, retry)

    def set_warm_white_255(self, level, persist=True, retry=2):
        self._bulb.set_warm_white_255(level, persist, retry)

    def set_cold_white(self, level, persist=True, retry=2):
        self._bulb.set_cold_white(level, persist, retry)

    def set_cold_white_255(self, level, persist=True, retry=2):
        self._bulb.set_cold_white_255(level, persist, retry)

    def set_white_temperature(self, temperature, brightness,
                              persist=True, retry=2):
        # Assume output temperature of between 2700 and 6500 Kelvin, and scale
        # the warm and cold LEDs linearly to provide that
        self._bulb.set_white_temperature(temperature, brightness, persist, retry)

    @property
    def warm_white(self):
        return self._bulb.warm_white

    @property
    def warm_white_255(self):
        return self._bulb.warm_white_255

    @property
    def cold_white(self):
        return self._bulb.cold_white


# Product models
# add more to this list when support is added
# refers to "model" key
MAGICHOME_BRAND_LOOKUP = {
    "AK001-ZJ210": MagicHomeRGBWBulb
}


def scan_magichome():
    scanner = BulbScanner()
    scanner.scan(timeout=2)
    devices = scanner.get_bulb_info()
    for d in devices:
        product_model = d.get("model", "magic_home")
        host = d.get("ipaddr")
        clazz = MAGICHOME_BRAND_LOOKUP.get(product_model) or MagicHomeDevice
        d["name"] = d.get("name") or clazz.__name__
        yield clazz(host, d["name"], d, product_model=product_model)


def get_device(ip):
    for device in scan_magichome():
        if device.host == ip:
            return device
    return None


if __name__ == '__main__':
    from pprint import pprint
    from time import sleep


    bulb = get_device('192.168.1.65')
    bulb.change_color("cyan")
    print(bulb.__dict__)

    bulb.set_warm_white_255(255)
    # bulb.color_cycle()
    # sleep(60)
    # bulb.reset()



import sys
import socket

import nmap
import bluetooth

from jarbas_house.devices import GenericDevice


# Get your local network IP address. e.g. 192.168.178.X
def get_lan_ip():
    try:
        return ([(s.connect(('8.8.8.8', 80)), s.getsockname()[0], s.close()) for s in
                 [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1])
    except socket.error as e:
        sys.stderr.write(str(e) + "\n")  # probably offline / no internet connection
        sys.exit(e.errno)


# Scan you local network for all hosts
def scan_wifi(nmap_args="-sn"):
    hosts = str(".".join(get_lan_ip().split(".")[:-1])) + ".0/24"
    scanner = nmap.PortScanner()
    scanner.scan(hosts=hosts, arguments=nmap_args)

    for ip in scanner.all_hosts():
        name = scanner[ip].hostname()
        yield GenericDevice(ip, name, scanner[ip])


def scan_bluetooth():
    for addr, name in bluetooth.discover_devices(lookup_names=True):
        yield GenericDevice(addr, name, {"address": addr, "name": name, "device_type": "bluetooth"})


if __name__ == "__main__":
    from pprint import pprint
    for host in scan_wifi():
        pprint(host.as_dict)
    for host in scan_bluetooth():
        pprint(host.as_dict)



from webcolors import name_to_hex, name_to_rgb, name_to_rgb_percent, \
    rgb_percent_to_hex, \
    rgb_percent_to_name, rgb_percent_to_rgb, rgb_to_hex, rgb_to_name, \
    rgb_to_rgb_percent, \
    hex_to_name, hex_to_rgb, hex_to_rgb_percent
from colorsys import rgb_to_yiq, yiq_to_rgb, rgb_to_hls, hls_to_rgb, \
    rgb_to_hsv, hsv_to_rgb


def hsv_to_name(h, s, v):
    rgb = hsv_to_rgb(h, s, v)
    return rgb_to_name(rgb)


def name_to_hsv(name):
    r, g, b = name_to_rgb(name)
    return rgb_to_hsv(r, g, b)


def hex_to_hsv(hex_color):
    r, g, b = hex_to_rgb(hex_color)
    h, s, v = rgb_to_hsv(r, g, b)
    return h, s, v


class Color:
    def __init__(self, r=255, g=255, b=255):
        self._r = r
        self._g = g
        self._b = b

    @staticmethod
    def from_name(name):
        return Color(*name_to_rgb(name))

    @staticmethod
    def from_hls(h, l, s):
        return Color(*hls_to_rgb(h, l, s))

    @staticmethod
    def from_yiq(y, i, q):
        return Color(*yiq_to_rgb(y, i, q))

    @staticmethod
    def from_rgb(r, g, b):
        return Color(r, g, b)

    @staticmethod
    def from_hsv(h, s, v):
        return Color(*hsv_to_rgb(h, s, v))

    @staticmethod
    def from_hex(hexcode):
        return Color(*hex_to_rgb(hexcode))

    @property
    def rgb(self):
        return (int(self._r), int(self._g), int(self._b))

    @property
    def yiq(self):
        return rgb_to_yiq(*self.rgb)

    @property
    def hls(self):
        return rgb_to_hls(*self.rgb)

    @property
    def hsv(self):
        return rgb_to_hsv(*self.rgb)

    @property
    def name(self):
        try:
            return rgb_to_name(self.rgb)
        except:
            return "unnamed color"

    @property
    def hex(self):
        return rgb_to_hex(self.rgb)

    def __repr__(self):
        return "{name}:{rgb}".format(name=self.name, rgb=self.rgb)

    @property
    def as_dict(self):
        return {
            "name": self.name,
            "rgb": self.rgb,
            "hsv": self.hsv,
            "hex": self.hex,
            "hls": self.hls,
            "yiq": self.yiq
        }


if __name__ == "__main__":
    hex_color = "#ff0000"
    color_name = hex_to_name(hex_color)
    print(color_name)
    h, s, v = name_to_hsv(color_name)
    print(h, s, v)
    r, g, b = hex_to_rgb(hex_color)
    print(r, g, b)
    h, s, v = rgb_to_hsv(r, g, b)
    print(h, s, v)

    color = Color()
    print(color.name)
    print(color.as_dict)
    color = Color.from_name("red")
    print(color.name, color.rgb)
    print(color.as_dict)
    color = Color.from_name("violet")
    print(color.name, color.rgb)

    # color = Color()
    print(color.as_dict)
